<?php

  // including required PHP files
  require_once 'config.php';           // configuration parameters
  require_once 'HOGTree.class.php';    // the Tree class
  require_once 'TreeParser.class.php'; // the TreeParser class

  // reading the species names table from file
  say( "Reading the bacterial strain names table ... " );
  $bacNames = readNamesTable( $config['NamesTableFile'] );
  say( "done (" . count($bacNames) . " bac.strains" . ")\n" );
  say( "Reading taxa and genera lists ... " );
  $taxa     = readTaxaOrder($config['TaxaAntiquFile']);
  $gen2tax  = readGeneraTaxa($config['TaxaGeneraFile']);
  say( "done (" . count($taxa) . " taxa, " . count($gen2tax) . " genera)\n" );

  // static data for HOGTree class
  HOGTree::$taxa    = $taxa;      // taxa list in order of antiquity
  HOGTree::$strains = $bacNames;  // bacterial strains abbreviations and full names
  HOGTree::$gen2tax = $gen2tax;   // genera to taxa mapping

  // Starting the main loop...
  say( "Starting the main loop.\n" );
  $treeFiles = makeTreeFilesList( $config['TreesDirectory'],
                                  $config['TreesExtension'] );

  file_put_contents( $config['ResultFileName'], "" );
  $nmin = $config['n']; // minimal acceptable number of bacterial leaves
  $tn = 0; // current tree number (for progress indication only)
  $tc = count( $treeFiles ); // total trees count (for progress indication only)
  $log_file_name = "nul";
  foreach ( $treeFiles as $treeName => $fileName )
  {
    ++$tn;
    say( "\n* Input file '$treeName' ($tn/$tc):\n" );

    log_start( "$config[DebugDirectory]/$treeName.log" );
    log_out( "Log file for '$treeName'.\n\n" );

    // setting bacterial strains list to normal state
    HOGTree::$strains = $bacNames;

    // getting a tree from current file
    log_out( "[STEP 0]\n" );
    log_out( "Reading and parsing input tree.\n\n" );
    say( "  Reading... ");
    $newickTree = file_get_contents( $fileName );
      // a text representation of the tree
    if ( $newickTree === false )
      die( "Can't read from file '$fileName'." );
    if ( empty( $newickTree ) )
    {
      say( " The file is empty.\n" );
      continue;
    }
    log_out( "Input tree:\n$newickTree\n\n" );
    say( "Parsing... " );
    $tree = parseTree( $newickTree );
    say( "Labels... " );
    $hogTree = new HOGTree( $tree ); // making a tree with parsed labels
    log_out( "Parsed input tree:\n" . $hogTree->toNewickCompact() . "\n\n" );
    $tree->destroy(); // freeing the memory...
    unset( $tree );
    say( "done\n" );
    // now we've got a fully parsed tree in $hogTree object

    // first we'll check the tree according to Step 1 of Req.Spec.
    log_out( "[STEP 1]\n" );
    $skip = false; // if $skip==true -- the tree will be ignored

    $nL   = $hogTree->numberOfLeaves();
    $nBac = $hogTree->numberOfBacLeaves();
    $tTax = $hogTree->taxaList();
    sort($tTax);
    $nTax = count($tTax);

    log_out( "Number of bacterial leaves: $nBac (of $nL total).\n" );
    log_out( "Present taxa: " . implode(', ', $tTax) . ".\n" );

    say( "  Bacterial leaves: $nBac (of $nL)" );
    if ( !$skip )
      if ( $nBac < $nmin ) // not anough bacterial leaves
      {
        say( " < $nmin" );
        $skip = true;
      }
    say( ". " );
    if ( !$skip )
    {
      say( "Major taxa: $nTax" );
      if ( $nTax < 3 ) // not anough taxa
      {
        say( " < 3" );
        $skip = true;
      }
      say( ". " );
    }
    if ( !$skip )
    {
      say( "Accepted for S1.\n" );
      log_out( "Accepted for S1.\n\n" );
    }
    else
    {
      say( "Skipped.\n" );
      log_out( "Skipped.\n\n" );
      continue;
    }

    // removing non-bacterial branches
    say( "  Removing non-bacterial branches... " );
    $hogTree->removeNonBacBranches();
    $hogTree = $hogTree->removeTransitNodes();
    log_out( "The tree without non-bacterial branches:\n" .
             $hogTree->toNewickCompact() . "\n\n" );
    say( "done\n" );

    // Beginning of Step 2 of Req.Spec.
    log_out( "[STEP 2]\n" );
    say( "  Removing edges with low bootstrap values... " );
    log_out( "Removing edges with low bootstrap values.\n\n" );

    $rtc = 0; // resulting trees count (for progress indication only)
    $forest = array(); // a temporary stack of current tree parts
    array_push( $forest, $hogTree ); // putting current tree on stack
    $s2trees = array(); // trees accepted to S2
    while ( count( $forest ) ) // while stack's not empty
    {
      $curTree = array_pop( $forest ); // getting a tree from the top of the stack
      if ( $curTree->numberOfLeaves() < $nmin ) // the tree is too small
      {
        $curTree->destroy();  // working hard to prevent memleaks
        continue;
      }
      $curTree = $curTree->removeTransitNodes(); // removing the "transit nodes"
      $bsList = $curTree->getBSList(); // getting the list of bootstrap values
      if ( !count( $bsList ) )
        continue; // no more edges have bootstrap values
      usort( $bsList, "cmp_bslist" ); // sorting bootstrap values
      list( $mbs, $mbsNode ) = $bsList[0]; // lowest BS value and it's node
      if ( $mbs < $config['b'] ) // if it's below the threshold...
      {
        $tParts = $mbsNode->separate(); // splitting the tree into two parts
        if ( $tParts === false )
          die( "Something bad happend while trying to split the tree. :(" );
        foreach ( $tParts as $tPart )
          array_push( $forest, $tPart ); // putting both parts on stack
      }
      else // otherwise (all BS values are above the threshold) ...
      {
        $s2trees[] = $curTree; // accepting the tree for S2
        ++ $rtc;
        log_out( "Accepted tree #$rtc:\n" . $curTree->toNewickCompact() . "\n\n" );
      }
    }

    say( "done\n" );
    say( "  Number of trees in S2: $rtc.\n" );
    // End of Step 2 of Req.Spec.

    // Beginning of Steps 3 & 4 of Req.Spec.
    say( "  Filtering out trees having p<$config[p] and defining roots.\n" );

    unset($hogTree);  // now working with $s2trees, forget about $hogTree
    foreach ( $s2trees as $s2tn => $curTree )
    {
      say( "  Tree #" . ($s2tn+1) . ": " );
      log_out( "Processing tree #" . ($s2tn+1) . "\n" );
      $tTax = $curTree->taxaList();
      sort($tTax);
      $oldestTaxon = null;
      $oldestTaxon2 = null;
      log_out( "Present taxa in order of antiquity (descending):\n" );
      foreach ( $taxa as $taxon => $antiquity )
        if ( in_array( $taxon, $tTax ) )
        {
          log_out( " * $taxon ($antiquity)\n" );
          if ( $oldestTaxon === null )
          {
            $oldestTaxon = $taxon;
            $oldestTaxon2 = null;
          }
          elseif ( $antiquity === $taxa[$oldestTaxon] )
            $oldestTaxon2 = $taxon;
        }
      if ( count($tTax) < 3 )
      {
        say("< 3 taxa, skipped.\n");
        log_out( "Too few taxa. Skiped.\n\n" );
        $curTree->destroy();  // working hard to prevent memleaks
        continue;
      }

      log_out( "\n" );

      // filtering by 'p' value
      // Step 3 of Req.Spec.
      log_out( "[STEP 3]\n" );
      $skip = false;
      log_out( "Defining 'd' and 'p' values.\n" );
      $oldestTaxons = array($oldestTaxon);
      if ( $oldestTaxon2 !== null )
        $oldestTaxons[] = $oldestTaxon2;
      list($dmt, $p) = find_oldest_edge($curTree, $oldestTaxons);
      if ( $p < $config['p'] )
      {
        log_out( "p < $config[p].\n\n" );
        $skip = true;
      }

      // Step 4 of Req.Spec.
      log_out( "[STEP 4]\n" );
      log_out( "Filtering by 'p' value(s).\n" );
      if ( $skip )
      { // removing the whole tree and skipping to the next one
        $curTree->destroy(); // freeing the memory...
        say( "removed.\n" );
        log_out( "Removed\n\n" );
        continue;
      }
      say( "accepted. " );
      log_out( "Accepted.\n\n" );
      // finished filtering by 'p' value

      $dmtp = $dmt->getParent();
      if ( $dmt->isRoot() ) // never happens (?)
        $resTree = $dmt;
      elseif ( $dmtp->isRoot() and $dmtp->childrenCount() == 2 )
        $resTree = $dmtp;
      else
      {
        $dmtp->unlinkChild($dmt);
        $dmtp->becomeRoot();
        $dmtp = $dmtp->removeTransitNodes();
        $resTree = new HOGTree();
        $resTree->addChild( $dmt );
        $resTree->addChild( $dmtp );
      }

      file_put_contents( $config['ResultFileName'],
                         $resTree->toNewickWithTaxa() . "\n",
                         FILE_APPEND );
      log_out( "Result tree:\n" . $resTree->toNewickWithTaxa() . "\n" );
      $resTree->destroy();  // working hard to prevent memleaks

      say( "\n" );
      log_out( "\n" );
    }  // loop through the trees in S2 for current input tree
  }  // main loop through the input trees list

  say( "\nAll done." );



  // sends the given text to... well, stdout, but can send it somewhere else :)
  function say( $text )
  {
    echo $text;
  }

  // debugging logs...
  function log_start( $file_name )
  {
    global $log_file_name;
    $log_file_name = $file_name;
    $result = @file_put_contents( $log_file_name, "" );
    if ( $result === false )
      $log_file_name = "nul";
  }
  function log_out( $text )
  {
    global $log_file_name;
    file_put_contents( $log_file_name, $text, FILE_APPEND );
  }

  // reads the species names table from the given file
  function readNamesTable( $filename )
  {
    $bacNamesFile = file( $filename );
    if ( $bacNamesFile === false )
      die( "Can't read names table file '$filename'." );
    if ( trim( $bacNamesFile[0] ) !== "CODE\tSPECIES NAME" )
      die( "Unexpected header line in names table file:\n$bacNamesFile[0]." );
    unset( $bacNamesFile[0] );
    $bacNames = array(); // mapping of abbreviated species names to their full names
    foreach ( $bacNamesFile as $row )
      if ( ! empty( $row ) )
      {
        list( $abbr, $name ) = explode( "\t", trim( $row ) );
        list( $cur_family ) = explode( " ", $name );
        $bacNames[$abbr] = $name;
      }
    asort( $bacNames );
    return $bacNames;
  }

  // reads the genera-taxa correspondence from the given file
  function readGeneraTaxa( $filename )
  {
    $taxaGeneraFile = file( $filename );
    if ( $taxaGeneraFile === false )
      die( "Can't read taxa-genera file '$filename'." );
    $genera = array();
    $taxon = null;
    foreach ( $taxaGeneraFile as $line )
    {
      $line = trim($line);
      if ( !strlen($line) )
        continue;
      list($genus, $taxon) = explode("\t", $line);
      $genera[$genus] = $taxon;
    }
    return $genera;
  }

  // reads the genera-taxa correspondence from the given file (older version)
  function readGeneraTaxa_v0( $filename )
  {
    $taxaGeneraFile = file( $filename );
    if ( $taxaGeneraFile === false )
      die( "Can't read taxa-genera file '$filename'." );
    #$taxa = array();
    $genera = array();
    $taxon = null;
    foreach ( $taxaGeneraFile as $line )
    {
      $line = trim($line);
      if ( !strlen($line) )
        continue;
      if ( $line[0] === '>' )
      {
        $taxon = substr($line, 1);
        continue;
      }
      $genera[$line] = $taxon;
      #$taxa[$taxon][] = $line;
    }
    return $genera;
  }

  // reads the taxa antiquity order from the given file
  function readTaxaOrder( $filename )
  {
    $taxaFile = file( $filename );
    if ( $taxaFile === false )
      die( "Can't read taxa order file '$filename'." );
    $taxa = array();
    foreach ( $taxaFile as $line )
    {
      $line = trim($line);
      if ( !strlen($line) )
        continue;
      if ( $line[0] === '#' )
        continue;
      list($antiquity, $taxon) = explode("\t", $line);
      $antiquity = intval($antiquity);
      $taxa[$taxon] = $antiquity;
    }
    asort($taxa);
    return $taxa;
  }

  // takes a directory name and file extension and returns
  // the list of matching filenames
  function makeTreeFilesList( $dir, $ext )
  {
    $treeFilesMask = "$dir/*.$ext";
    $treeFilesList = glob( $treeFilesMask );
    if ( empty( $treeFilesList ) )
      die( "No files found matching mask '$treeFilesMask'." );
    $treeFiles = array();
    foreach ( $treeFilesList as $fileName )
    {
      $treeName = preg_replace( "~\.$ext$~", "", basename( $fileName ) );
      $treeFiles[$treeName] = $fileName;
    }
    return $treeFiles;
  }

  // parses the given newick tree and returns a Tree object
  function parseTree( $newickTree )
  {
    $tParser = new TreeParser( $newickTree ); // parsing the tree
    $trees = $tParser->getTrees(); // getting the result
    if ( $tParser->errorOccured() )
      die( "Tree parser error: \"" . $tParser->getErrorMessage() . "\".\n" );
    if ( empty( $trees ) )
      die( "No trees." );

    // destroying trees except the 1st (if there were many)
    $first = true;
    foreach ( $trees as $tree )
      if ( $first )
        $first = false;
      else
        $tree->destroy();

    return $trees[0];
  }

  // finds the oldest edge in the tree according to the oldest taxon
  function find_oldest_edge($curTree, $oldestTaxons)
  {
    // setting bacterial strains list to reduced state (oldest taxon excluded)
    $strains = HOGTree::$strains;
    HOGTree::$strains = excludeTaxons( $strains, $oldestTaxons );
    $curTree->updateStats();  // updating cache

    $curTree->calcD_rec(); // calculating 'd' values in the whole tree
    //file_put_contents( "out.tre", $curTree->toNewickDebug() );
    $dList = $curTree->getDList(); // getting the list of 'd' values
    usort( $dList, "cmp_dlist" );  // sorting it in the reverse order
    list( $dm, $dmt ) = $dList[0]; // 'd_max' from Req.Spec
    list( $dpm )      = $dList[1]; // 'd_predmax' from Req.Spec
    $p = sqrt($dm) - sqrt($dpm) + $dm*$dm;  // 'p' from Req.Spec
    say( "p=" . number_format($p,4) . ", " );
    log_out( "chosen taxon(s): " . implode(', ', $oldestTaxons) . "\n" );
    log_out( "d_max = $dm\n" );
    log_out( "d_predmax = $dpm\n" );
    log_out( "p = $p\n" );
    log_out( "d_max tree:\n" . $dmt->toNewickCompact() . "\n" );
    log_out( "debug tree:\n" . $curTree->toNewickDebug() . "\n\n" );

    // setting bacterial strains list to normal state
    HOGTree::$strains = $strains;
    $curTree->updateStats();  // updating cache

    return array($dmt, $p);
  }

  // excludes pseudo-non-bacterial taxon from bacterial strains list
  function excludeTaxons( $bacNames, $taxons )
  {
    global $gen2tax;
    $bacNames_ = array();
    foreach ( $bacNames as $id => $name )
    {
      list( $cur_genus ) = explode( " ", $name );
      if ( !isset($gen2tax[$cur_genus]) )
        exit("Unknown genus: '$cur_genus'.");
      if ( !in_array($gen2tax[$cur_genus], $taxons) )
        $bacNames_[$id] = $name;
    }
    return $bacNames_;
  }

  // user callback function for reverse sort by 1st array element
  function cmp_dlist($a, $b)
  {
    if ($a[0] == $b[0])
        return 0;
    return ($a[0] > $b[0]) ? -1 : 1;
  }

  // user callback function for sort by 1st array element
  function cmp_bslist($a, $b)
  {
    if ($a[0] == $b[0])
        return 0;
    return ($a[0] < $b[0]) ? -1 : 1;
  }

?>
